import axinite as ax
import numpy as np
import astropy.units as u
from astropy.coordinates import CartesianRepresentation
from numba import jit

def load(delta, limit, backend, *bodies, t=0 * u.s, modifier=None, action=None, action_frequency=200):
    """Loads a simulation from a backend.

    Args:
        delta (u.Quantity): The change in time between each step.
        limit (u.Quantity): The limit of the simulation.
        backend (function): The backend to load the simulation from.
        t (u.Quantity, optional): The initial time. Defaults to 0 * u.s.
        modifier (function, optional): The modifier to apply to forces in the simulation. Defaults to None.
        action (function, optional): An action to call with frequency `action_frequency`. Defaults to None.
        action_frequency (int, optional): The frequency at which to call the `action`. Defaults to 200.

    Returns:
        np.ndarray: An array of bodies.
    """
    _bodies = backend(delta, limit, ax.get_inner_bodies(bodies), action=action, modifier=modifier, t=t, action_frequency=action_frequency)
    __bodies = []
    for body in _bodies:
        _body = ax.Body(body["n"], body["m"], limit, delta)
        _body._inner = body
        __bodies.append(_body)
    return __bodies



import axinite as ax
import astropy.units as u
import numpy as np
class Body:
    "A class that represents a body in the simulation."
    def __init__(self, name: str, mass: np.float64, limit: np.float64, delta: np.float64, position: np.ndarray = None, velocity: np.ndarray = None):
        """Initializes a new Body object.

        Args:
            name (str): The body's name.
            mass (u.Quantity): The mass of the body in kilograms.
            position (CartesianRepresentation): The initial position of the body (in vector form).
            velocity (CartesianRepresentation): The initial velocity of the body (in vector form).
            radius (u.Quantity): The radius of the body in meters.
            color (str, optional): The color of the body. Defaults to "", telling the frontend to choose from a list.
            light (bool, optional): Whether the body should give of light. Defaults to False.
            retain (_type_, optional): How many points the body should retain on it's trail. Defaults to None, meaning "go with default".
            radius_multiplier (int, optional): A multiplier to be applied to the radius. Defaults to 1.
        """

        self.mass: np.float64 = mass
        "The mass of the body in kilograms."

        self.name: str = name
        "The name of the body."

        self._inner = ax._body(limit, delta, name, mass)

        if position is not None: self._inner["r"][0] = position
        if velocity is not None: self._inner["v"][0] = velocity

        self._inner["n"] = name
        self._inner["m"] = mass

        self.radius: np.float64 = -1
        "The radius of the body in meters."

        self.color: str = ""
        "The color of the body."

        self.light: bool = False
        "Whether the body should give off light."

        self.retain: int = None
        "How many points the body should retain on it's trail."

        self.radius_multiplier: int = 1
        "A multiplier which was applied to the radius."

    def r(self, t: np.float64) -> np.ndarray:
        """Returns the position of the body at a specific time.

        Args:
            t (np.float64): The time to get the position at.

        Returns:
            np.ndarray: The position of the body at the time.
        """
        return self._inner["r"][int(t)]
    
    def v(self, t: np.float64) -> np.ndarray:
        """Returns the velocity of the body at a specific time.

        Args:
            t (np.float64): The time to get the velocity at.

        Returns:
            np.ndarray: The velocity of the body at the time.
        """
        return self._inner["v"][int(t)]


import math
import numpy as np
import axinite as ax
from numba import njit

G = 6.67430e-11
def body_dtype(limit, delta): return np.dtype([
    ("n", "U20"),
    ("m", np.float64),
    ("r", np.float64, (int(limit/delta), 3)),
    ("v", np.float64, (int(limit/delta), 3))
])
def _body(limit, delta, name, mass): 
    return np.array((name, mass, np.zeros((int(limit/delta), 3)), np.zeros((int(limit/delta), 3))), dtype=ax.body_dtype(limit, delta))

def get_inner_bodies(bodies):
    _bodies = ()
    for body in bodies: _bodies += (body._inner,)
    return _bodies

def create_outer_bodies(bodies, limit, delta):
    _bodies = []
    for body in bodies:
        _body = ax.Body(str(body["n"]), body["m"], limit, delta)
        _body._inner = body
        _bodies.append(_body)
    return _bodies

@njit 
def vector_magnitude_jit(vec: np.ndarray) -> float:
    """Calculates the magnitude of a vector.

    Args:
        vec (np.ndarray): The vector to calculate the magnitude of.

    Returns:
        float: The magnitude of the vector.
    """
    return np.sqrt(np.sum(vec**2))

@njit
def unit_vector_jit(vec: np.ndarray) -> np.ndarray:
    """Calculates the unit vector of a vector.

    Args:
        vec (np.ndarray): The vector to calculate the unit vector of.

    Returns:
        np.ndarray: The unit vector of the vector.
    """
    mag = vector_magnitude_jit(vec)
    return vec / mag if mag > 0 else vec

@njit
def gravitational_force_jit(m1: np.float64, m2: np.float64, r: np.ndarray) -> np.ndarray:
    """Calculates the gravitational force between two bodies.

    Args:
        m1 (u.Quantity): The mass of the first body.
        m2 (u.Quantity): The mass of the second body.
        r (np.ndarray): The vector between the two bodies.

    Returns:
        np.ndarray: The gravitational force between the two bodies.
    """
    mag = vector_magnitude_jit(r)
    if mag == 0:
        return np.zeros(3)
    return -G *((m1 * m2) / mag**2) * unit_vector_jit(r)


import axinite.tools as axtools
import json

def combine(meta: dict, file: dict, indent=4) -> str:
    """Combines a .meta.ax file and a .ax/.tmpl.ax file

    Args:
        meta (dict): A dictionary of the .meta.ax file
        file (dict): A dictionary of the .ax/.tmpl.ax file

    Returns:
        str: A JSON string of the combined data
    """
    metadata = json.loads(meta)
    filedata = json.loads(file)
    if "name" in metadata and "name" in filedata: del filedata['name']
    if "author" in metadata and "author" in filedata: del filedata['author']
    combined = {**metadata, **filedata}
    return json.dumps(combined, indent=indent)


import axinite.tools as axtools
import json

def save(args: axtools.AxiniteArgs, path: str):
    with open(path, 'w+') as f:
        data = {
            "name": args.name,
            "delta": float(args.delta),
            "limit": float(args.limit),
            "t": float(args.t),
            "bodies": []
        }

        for body in args.bodies: 
            body_data = {
                "name": str(body.name),
                "mass": float(body.mass),
                "radius": float(body.radius),
                "r": [],
                "v": []
            }

            for i, r in enumerate(body._inner["r"]):
                _r = [float(v) for v in r]
                body_data["r"].append(_r)
            for i, v in enumerate(body._inner["v"]):
                _v = [float(v) for v in v]
                body_data["v"].append(_v)

            if body.color != None:
                body_data["color"] = body.color
            if body.retain != None:
                body_data["retain"] = int(body.retain)
            if body.light != None:
                body_data["light"] = body.light

            data["bodies"].append(body_data)

        if args.radius_multiplier is not None:
            data["radius_multiplier"] = int(args.radius_multiplier)

        if args.rate is not None:
            data["rate"] = args.rate

        if args.retain is not None:
            data["retain"] = args.retain
        
        if args.frontend_args != {}:
            data["frontend_args"] = args.frontend_args

        json.dump(data, f, indent=4)


import axinite.tools as axtools
import axinite as ax

def live(args: axtools.AxiniteArgs, frontend: 'function') -> None:
    """Watch a preloaded simulation live.

    Args:
        args (axtools.AxiniteArgs): The arguments for the simulation.
        frontend (function): The frontend to use.
    """
    if args.rate is None:
        args.rate = 100
    if args.radius_multiplier is None:
        args.radius_multiplier = 1
    if args.retain is None:
       args.retain = 200

    t = 0
    while t < args.limit:
        frontend[0](args.bodies, t, limit=args.limit, delta=args.delta, n=int(t / args.delta))
        t += args.delta
    frontend[1]()


import numpy as np
import axinite as ax

class Body:
    def __init__(self, name: str, mass: np.float64, limit: np.float64, delta: np.float64, position: np.ndarray = None, velocity: np.ndarray = None):
        """Initializes a new Body object.

        Args:
            mass (u.Quantity): Mass of the object in kilograms.
            position (CartesianRepresentation): The initial position of the object.
            velocity (CartesianRepresentation): The initial velocity of the object.
        """
        self.mass = mass
        "The mass of the object in kilograms."

        self.name = name
        "The name of the object."

        self._inner = ax._body(limit, delta, name, mass)

        if position is not None: self._inner["r"][0] = position
        if velocity is not None: self._inner["v"][0] = velocity

        self._inner["n"] = name
        self._inner["m"] = mass

    def r(self, t: np.float64) -> np.ndarray:
        """Returns the position of the object at a specific time.

        Args:
            t (np.float64): The time to get the position at.

        Returns:
            np.ndarray: The position of the object at the time.
        """
        return self._inner["r"][int(t)]

    def v(self, t: np.float64) -> np.ndarray:
        """Returns the velocity of the object at a specific time.

        Args:
            t (np.float64): The time to get the velocity at.

        Returns:
            np.ndarray: The velocity of the object at the time.
        """
        return self._inner["v"][int(t)]


from axinite.tools.args import AxiniteArgs
from axinite.tools.body import Body
from axinite.tools.functions import interpret_time, data_to_body, interpret_distance, interpret_mass, \
    string_to_color, create_sphere, max_axis_length, min_axis_length, from_body
from axinite.tools.load import load
from axinite.tools.read import read
from axinite.tools.show import show
from axinite.tools.live import live
from axinite.tools.run import run
from axinite.tools.frontends.vpython import vpython_frontend
from axinite.tools.frontends.plotly import plotly_frontend
from axinite.tools.save import save
from axinite.tools.combine import combine


"The `axinite` module is the main module for the Axinite celestial mechanics engine."

from axinite.body import Body
from axinite.functions import vector_magnitude_jit, unit_vector_jit, gravitational_force_jit, body_dtype, \
    get_inner_bodies, _body, create_outer_bodies
import axinite.functions as functions
from axinite.load import load
from axinite.backends.euler import euler_backend, euler_nojit_backend
from axinite.backends.verlet import verlet_backend, verlet_nojit_backend